from numpy import ndarray, float64, int32, zeros, dtype
from . import _cpphelpers as lib
from typing import Tuple, Sequence
from .utils import _sanitize_subset
from delayedarray import is_sparse, extract_dense_array, extract_sparse_array, is_masked, chunk_grid, SimpleGrid, chunk_shape_to_grid

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def _factorize(group):
    mapping = {}
    indices = ndarray((len(group),), int32)
    levels = []
    for i, x in enumerate(group):
        if x not in mapping:
            mapping[x] = len(levels)
            levels.append(x)
        indices[i] = mapping[x]
    return levels, indices


class TatamiNumericPointer:
    """Pointer to a tatami numeric matrix allocated by C++ code. Instances of
    this class should only be created by developers and used within package
    functions; this is done by fetching the :py:attr:`~ptr` attribute and
    casting it to a ``Mattress*`` in C++ code. Pointers are expected to be
    transient within a Python session; they should not be serialized, nor
    should they be visible to end users. Each instance will automatically free
    the C++-allocated memory upon its own destruction.
    """

    def __init__(self, ptr: int, obj: list):
        self.ptr = ptr
        self.obj = obj

    def __del__(self):
        lib.free_mat(self.ptr)

    def nrow(self) -> int:
        """Get number of rows.

        Returns:
            Number of rows.
        """
        return lib.extract_nrow(self.ptr)

    def ncol(self) -> int:
        """Get number of columns.

        Returns:
            Number of columns.
        """
        return lib.extract_ncol(self.ptr)

    @property
    def shape(self) -> Tuple[int, int]:
        """Shape of the matrix, to masquerade as a NumPy-like object."""
        return (self.nrow(), self.ncol())

    @property
    def dtype(self) -> dtype:
        """Type of the matrix, to masquerade as a NumPy-like object."""
        return dtype("float64")

    def __array__(self) -> ndarray:
        """Realize the underlying matrix into a dense NumPy array.

        Returns:
            Contents of the underlying matrix.
        """
        output = ndarray(self.shape, dtype=float64, order="C")
        lib.extract_dense_full(self.ptr, output.ctypes.data)
        return output

    def __generic_extract__(self, subset: Tuple[Sequence[int], ...]) -> ndarray:
        """Enable DelayedArray extraction of subsets of data from the matrix.

        See :py:meth:`~delayedarray.DelayedArray.DelayedArray.__DelayedArray_extract__` for details.

        Returns:
            Contents of the underlying matrix.
        """
        rfull = self.nrow()
        rnoop, rsub = _sanitize_subset(subset[0], rfull)
        roffset = 0
        if rnoop:
            rlen = rfull
        else:
            roffset = rsub.ctypes.data
            rlen = len(rsub)

        cfull = self.ncol()
        cnoop, csub = _sanitize_subset(subset[1], cfull)
        coffset = 0
        if cnoop:
            clen = cfull
        else:
            coffset = csub.ctypes.data
            clen = len(csub)

        output = ndarray((rlen, clen), dtype=float64)
        lib.extract_dense_subset(
            self.ptr, rnoop, roffset, rlen, cnoop, coffset, clen, output.ctypes.data
        )
        return output

    def __DelayedArray_dask__(self) -> ndarray:
        """Enable the use of the poiners with dask.

        See :py:meth:`~delayedarray.DelayedArray.DelayedArray.__DelayedArray_dask__` for details.

        This is largely a placeholder method for compatibility; it just realizes the underlying matrix into a dense
        array under the hood.

        Returns:
            Contents of the underlying matrix.
        """
        return self.__array___()

    def sparse(self) -> bool:
        """Is the matrix sparse?

        Returns:
            True if matrix is sparse.
        """
        return lib.extract_sparse(self.ptr) > 0

    def row(self, r: int) -> ndarray:
        """Access a row from the tatami matrix. This method is primarily intended for troubleshooting and should not be
        used to iterate over the matrix in production code. (Do that in C++ instead.)

        Args:
            r: Row to access.

        Returns:
            Row from the matrix. This is always in double-precision,
            regardless of the underlying representation.
        """
        output = ndarray((self.ncol(),), dtype="float64")
        lib.extract_row(self.ptr, r, output.ctypes.data)
        return output

    def column(self, c: int) -> ndarray:
        """Access a column from the tatami matrix. This method is primarily intended for troubleshooting and should not
        be used to iterate over the matrix in production code. (Do that in C++ instead.)

        Args:
            c: Column to access.

        Returns:
            Column from the matrix. This is always in double-precisino,
            regardless of the underlying representation.
        """
        output = ndarray((self.nrow(),), dtype="float64")
        lib.extract_column(self.ptr, c, output.ctypes.data)
        return output

    def row_sums(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute row sums.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row sums.
        """
        output = zeros((self.nrow(),), dtype=float64)
        lib.compute_row_sums(self.ptr, output.ctypes.data, num_threads)
        return output

    def column_sums(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute column sums.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column sums.
        """
        output = zeros((self.ncol(),), dtype=float64)
        lib.compute_column_sums(self.ptr, output.ctypes.data, num_threads)
        return output

    def row_variances(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute row variances.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row variances.
        """
        output = zeros((self.nrow(),), dtype=float64)
        lib.compute_row_variances(self.ptr, output.ctypes.data, num_threads)
        return output

    def column_variances(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute column variances.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column variances.
        """
        output = zeros((self.ncol(),), dtype=float64)
        lib.compute_column_variances(self.ptr, output.ctypes.data, num_threads)
        return output

    def row_medians(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute row medians.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row medians.
        """
        output = ndarray((self.nrow(),), dtype=float64)
        lib.compute_row_medians(self.ptr, output.ctypes.data, num_threads)
        return output

    def column_medians(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute column medians.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column medians.
        """
        output = ndarray((self.ncol(),), dtype=float64)
        lib.compute_column_medians(self.ptr, output.ctypes.data, num_threads)
        return output

    def row_mins(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute row minima.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row minima.
        """
        output = ndarray((self.nrow(),), dtype=float64)
        lib.compute_row_mins(self.ptr, output.ctypes.data, num_threads)
        return output

    def column_mins(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute column minima.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column mins.
        """
        output = ndarray((self.ncol(),), dtype=float64)
        lib.compute_column_mins(self.ptr, output.ctypes.data, num_threads)
        return output

    def row_maxs(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute row maxima.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row maxima.
        """
        output = ndarray((self.nrow(),), dtype=float64)
        lib.compute_row_maxs(self.ptr, output.ctypes.data, num_threads)
        return output

    def column_maxs(self, num_threads: int = 1) -> ndarray:
        """Convenience method to compute column maxima.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column maxs.
        """
        output = ndarray((self.ncol(),), dtype=float64)
        lib.compute_column_maxs(self.ptr, output.ctypes.data, num_threads)
        return output

    def row_ranges(self, num_threads: int = 1) -> Tuple[ndarray, ndarray]:
        """Convenience method to compute row ranges.

        Args:
            num_threads: Number of threads.

        Returns:
            Tuple containing the row minima and maxima.
        """
        min_output = ndarray((self.nrow(),), dtype=float64)
        max_output = ndarray((self.nrow(),), dtype=float64)
        lib.compute_row_ranges(
            self.ptr, min_output.ctypes.data, max_output.ctypes.data, num_threads
        )
        return (min_output, max_output)

    def column_ranges(self, num_threads: int = 1) -> Tuple[ndarray, ndarray]:
        """Convenience method to compute column ranges.

        Args:
            num_threads: Number of threads.

        Returns:
            Tuple containing the column minima and maxima.
        """
        min_output = ndarray((self.ncol(),), dtype=float64)
        max_output = ndarray((self.ncol(),), dtype=float64)
        lib.compute_column_ranges(
            self.ptr, min_output.ctypes.data, max_output.ctypes.data, num_threads
        )
        return (min_output, max_output)

    def row_nan_counts(self, num_threads: int = 1) -> ndarray:
        """Convenience method to count the number of NaNs on each row.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row NaN counts.
        """
        output = ndarray((self.nrow(),), dtype=int32)
        lib.compute_row_nan_counts(self.ptr, output.ctypes.data, num_threads)
        return output

    def column_nan_counts(self, num_threads: int = 1) -> ndarray:
        """Convenience method to count the number of NaNs on each column.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column NaN counts.
        """
        output = ndarray((self.ncol(),), dtype=int32)
        lib.compute_column_nan_counts(self.ptr, output.ctypes.data, num_threads)
        return output

    def row_medians_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[ndarray, list]:
        """Convenience method to compute the row-wise median for each group of columns.

        Args:
            group: Sequence of length equal to the number of columns of the matrix,
                containing the group assignment for each column.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each column represents
            a group and contains the row-wise medians for that group, across all rows of the matrix;
            and a list containing the unique levels of ``group`` represented by each column.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.ncol():
            raise ValueError(
                "'group' should have length equal to the number of columns"
            )

        output = ndarray((self.nrow(), len(lev)), dtype=float64)
        lib.compute_row_medians_by_group(
            self.ptr, ind.ctypes.data, output.ctypes.data, num_threads
        )
        return output, lev

    def column_medians_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[ndarray, list]:
        """Convenience method to compute the column-wise median for each group of row.

        Args:
            group: Sequence of length equal to the number of row of the matrix,
                containing the group assignment for each row.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each row represents a
            group and contains the column-wise medians for that group, across
            all columns of the matrix; and a list containing the unique levels
            of ``group`` represented by each row.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.nrow():
            raise ValueError("'group' should have length equal to the number of rows")

        output = ndarray((self.ncol(), len(lev)), dtype=float64)
        lib.compute_column_medians_by_group(
            self.ptr, ind.ctypes.data, output.ctypes.data, num_threads
        )
        return output.T, lev

    def row_sums_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[ndarray, list]:
        """Convenience method to compute the row-wise median for each group of columns.

        Args:
            group: Sequence of length equal to the number of columns of the matrix,
                containing the group assignment for each column.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each column represents
            a group and contains the row-wise sums for that group, across all
            rows of the matrix; and a list containing the unique levels of
            ``group`` represented by each column.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.ncol():
            raise ValueError(
                "'group' should have length equal to the number of columns"
            )

        output = ndarray((self.nrow(), len(lev)), dtype=float64)
        lib.compute_row_sums_by_group(
            self.ptr, ind.ctypes.data, output.ctypes.data, num_threads
        )
        return output, lev

    def column_sums_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[ndarray, list]:
        """Convenience method to compute the column-wise median for each group of row.

        Args:
            group: Sequence of length equal to the number of row of the matrix,
                containing the group assignment for each row.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each row represents a
            group and contains the column-wise sums for that group, across all
            columns of the matrix; and a list containing the unique levels of
            ``group`` represented by each row.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.nrow():
            raise ValueError("'group' should have length equal to the number of rows")

        output = ndarray((self.ncol(), len(lev)), dtype=float64)
        lib.compute_column_sums_by_group(
            self.ptr, ind.ctypes.data, output.ctypes.data, num_threads
        )
        return output.T, lev


@is_sparse.register
def is_sparse_tatami(x: TatamiNumericPointer):
    return x.sparse()


@extract_dense_array.register
def extract_dense_array_tatami(x: TatamiNumericPointer, subset: Tuple[Sequence[int], ...]) -> TatamiNumericPointer:
    """See :py:meth:`~delayedarray.extract_dense_array.extract_dense_array`."""
    return x.__generic_extract__(subset=subset)


@extract_sparse_array.register
def extract_sparse_array_tatami(x: TatamiNumericPointer, subset: Tuple[Sequence[int], ...]) -> TatamiNumericPointer:
    """See :py:meth:`~delayedarray.extract_sparse_array.extract_sparse_array`."""
    return x.__generic_extract__(subset=subset)


@is_masked.register
def is_masked_tatami(x: TatamiNumericPointer) -> bool:
    """See :py:meth:`~delayedarray.is_masked.is_masked`."""
    return False

@chunk_grid.register
def chunk_grid_tatami(x: TatamiNumericPointer) -> bool:
    """See :py:meth:`~delayedarray.chunk_grid.chunk_grid`."""
    return chunk_shape_to_grid((1, 1), x.shape, cost_factor=1)



import os
from typing import Union, Tuple, Sequence, List
import assorthead
import inspect
import numpy as np

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


def includes() -> List[str]:
    """Provides access to C++ headers (including tatami) for downstream packages.

    Returns:
        List of paths to the header files.
    """
    dirname = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    return [
        assorthead.includes(),
        os.path.join(dirname, "include"),
    ]


def _sanitize_subset(
    subset: Sequence[int], full: int
) -> Tuple[bool, Union[np.ndarray, None]]:
    is_noop = True
    if len(subset) == full:
        for i, x in enumerate(subset):
            if i != x:
                is_noop = False
                break
    else:
        is_noop = False

    if is_noop:
        return True, None

    if not isinstance(subset, np.ndarray):
        subset = np.array(subset, dtype=np.int32)
    else:
        subset = subset.astype(
            np.int32, copy=not (subset.flags.c_contiguous or subset.flags.f_contiguous)
        )

    return False, subset



from functools import singledispatch
from typing import Any

import numpy as np
import delayedarray
from biocutils.package_utils import is_package_installed


from .TatamiNumericPointer import TatamiNumericPointer
from . import _cpphelpers as lib
from .utils import _sanitize_subset

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


@singledispatch
def tatamize(x: Any) -> TatamiNumericPointer:
    """Converts Python matrix representations to a tatami pointer.

    Args:
        x: Any matrix-like object.

    Raises:
        NotImplementedError: if x is not supported.

    Returns:
        A pointer to tatami object.
    """
    raise NotImplementedError(
        f"tatamize is not supported for objects of class: {type(x)}"
    )


@tatamize.register
def _tatamize_pointer(x: TatamiNumericPointer) -> TatamiNumericPointer:
    return x  # no-op


@tatamize.register
def _tatamize_numpy(x: np.ndarray) -> TatamiNumericPointer:
    if len(x.shape) != 2:
        raise ValueError("'x' should be a 2-dimensional array")

    byrow = None
    if x.flags["C_CONTIGUOUS"]:
        byrow = True
    elif x.flags["F_CONTIGUOUS"]:
        byrow = False
    else:
        # I don't think it's possible to hit this, as a (non-view) ndarray
        # should be contiguous in at least one direction.
        raise ValueError("'x' must have contiguous storage for its arrays")

    return TatamiNumericPointer(
        ptr=lib.initialize_dense_matrix(
            x.shape[0],
            x.shape[1],
            str(x.dtype).encode("UTF-8"),
            x.ctypes.data,
            byrow,
        ),
        obj=[x],
    )

if is_package_installed("scipy"):
    import scipy.sparse

    @tatamize.register
    def _tatamize_sparse_csr_array(x: scipy.sparse.csr_array) -> TatamiNumericPointer:
        tmp = x.indptr.astype(np.uint64, copy=False)
        return TatamiNumericPointer(
            ptr=lib.initialize_compressed_sparse_matrix(
                x.shape[0],
                x.shape[1],
                len(x.data),
                str(x.data.dtype).encode("UTF-8"),
                x.data.ctypes.data,
                str(x.indices.dtype).encode("UTF-8"),
                x.indices.ctypes.data,
                tmp.ctypes.data,
                True,
            ),
            obj=[tmp, x],
        )


    @tatamize.register
    def _tatamize_sparse_csr_matrix(x: scipy.sparse.csr_matrix) -> TatamiNumericPointer:
        return _tatamize_sparse_csr_array(x)


    @tatamize.register
    def _tatamize_sparse_csc_array(x: scipy.sparse.csc_array) -> TatamiNumericPointer:
        tmp = x.indptr.astype(np.uint64, copy=False)
        return TatamiNumericPointer(
            ptr=lib.initialize_compressed_sparse_matrix(
                x.shape[0],
                x.shape[1],
                len(x.data),
                str(x.data.dtype).encode("UTF-8"),
                x.data.ctypes.data,
                str(x.indices.dtype).encode("UTF-8"),
                x.indices.ctypes.data,
                tmp.ctypes.data,
                False,
            ),
            obj=[tmp, x],
        )


    @tatamize.register
    def _tatamize_sparse_csc_matrix(x: scipy.sparse.csc_matrix) -> TatamiNumericPointer:
        return _tatamize_sparse_csc_array(x)


@tatamize.register
def _tatamize_delayed_array(x: delayedarray.DelayedArray) -> TatamiNumericPointer:
    return tatamize(x.seed)


@tatamize.register
def _tatamize_delayed_unary_isometric_op_simple(
    x: delayedarray.UnaryIsometricOpSimple,
) -> TatamiNumericPointer:
    components = tatamize(x.seed)
    ptr = lib.initialize_delayed_unary_isometric_op_simple(
        components.ptr, x.operation.encode("UTF-8")
    )
    return TatamiNumericPointer(ptr, components.obj)


@tatamize.register
def _tatamize_delayed_unary_isometric_op_with_args(
    x: delayedarray.UnaryIsometricOpWithArgs,
) -> TatamiNumericPointer:
    components = tatamize(x.seed)
    obj = components.obj

    if isinstance(x.value, np.ndarray):
        contents = x.value.astype(np.float64, copy=False)
        ptr = lib.initialize_delayed_unary_isometric_op_with_vector(
            components.ptr, x.operation.encode("UTF-8"), x.right, x.along, contents
        )
        obj.append(contents)
    else:
        ptr = lib.initialize_delayed_unary_isometric_op_with_scalar(
            components.ptr, x.operation.encode("UTF-8"), x.right, x.value
        )

    return TatamiNumericPointer(ptr, obj)


@tatamize.register
def _tatamize_delayed_subset(
    x: delayedarray.Subset,
) -> TatamiNumericPointer:
    components = tatamize(x.seed)
    obj = components.obj

    for dim in range(2):
        current = x.subset[dim]
        noop, current = _sanitize_subset(current, x.shape[dim])
        if not noop:
            ptr = lib.initialize_delayed_subset(
                components.ptr, dim, current, len(current)
            )
            obj.append(current)
            components = TatamiNumericPointer(ptr, obj)

    return components


@tatamize.register
def _tatamize_delayed_combine(
    x: delayedarray.Combine,
) -> TatamiNumericPointer:
    nseeds = len(x.seeds)
    objects = []
    converted = []
    ptrs = np.ndarray(nseeds, dtype=np.uintp)

    for i in range(nseeds):
        components = tatamize(x.seeds[i])
        converted.append(components)
        ptrs[i] = components.ptr
        objects += components.obj

    ptr = lib.initialize_delayed_combine(nseeds, ptrs.ctypes.data, x.along)
    return TatamiNumericPointer(ptr, objects)


@tatamize.register
def _tatamize_delayed_transpose(
    x: delayedarray.Transpose,
) -> TatamiNumericPointer:
    components = tatamize(x.seed)

    if x.perm == (1, 0):
        ptr = lib.initialize_delayed_transpose(components.ptr)
        components = TatamiNumericPointer(ptr, components.obj)

    return components


@tatamize.register
def _tatamize_delayed_binary_isometric_op(
    x: delayedarray.BinaryIsometricOp,
) -> TatamiNumericPointer:
    lcomponents = tatamize(x.left)
    rcomponents = tatamize(x.right)

    ptr = lib.initialize_delayed_binary_isometric_op(
        lcomponents.ptr, rcomponents.ptr, x.operation.encode("UTF-8")
    )

    return TatamiNumericPointer(ptr, lcomponents.obj + rcomponents.obj)


@tatamize.register
def _tatamize_delayed_round(
    x: delayedarray.Round,
) -> TatamiNumericPointer:
    components = tatamize(x.seed)

    if x.decimals != 0:
        raise NotImplementedError(
            "non-zero decimals in 'delayedarray.Round' are not yet supported"
        )

    ptr = lib.initialize_delayed_unary_isometric_op_simple(
        components.ptr, "round".encode("UTF-8")
    )

    return TatamiNumericPointer(ptr, components.obj)
